/*
 * sar, sadc, mpstat and iostat common routines.
 * (C) 1999-2004 by Sebastien GODARD (sysstat <at> wanadoo.fr)
 *
 ***************************************************************************
 * This program is free software; you can redistribute it and/or modify it *
 * under the terms of the GNU General Public License as published  by  the *
 * Free Software Foundation; either version 2 of the License, or (at  your *
 * option) any later version.                                              *
 *                                                                         *
 * This program is distributed in the hope that it  will  be  useful,  but *
 * WITHOUT ANY WARRANTY; without the implied warranty  of  MERCHANTABILITY *
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License *
 * for more details.                                                       *
 *                                                                         *
 * You should have received a copy of the GNU General Public License along *
 * with this program; if not, write to the Free Software Foundation, Inc., *
 * 675 Mass Ave, Cambridge, MA 02139, USA.                                 *
 ***************************************************************************
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>
#include <unistd.h>	/* For STDOUT_FILENO, among others */
#include <dirent.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/param.h>	/* for HZ */

/*
 * For PAGE_SIZE (which may be itself a call to getpagesize()).
 * PAGE_SHIFT no longer necessarily exists in <asm/page.h>. So
 * we use PAGE_SIZE to compute PAGE_SHIFT...
 */
#include <asm/page.h>

#include "common.h"

#ifdef USE_NLS
#include <locale.h>
#include <libintl.h>
#define _(string) gettext(string)
#else
#define _(string) (string)
#endif


/*
 ***************************************************************************
 * Get current date or time
 ***************************************************************************
 */
time_t get_localtime(struct tm *loc_time)
{
   time_t timer;
   struct tm *ltm;

   time(&timer);
   ltm = localtime(&timer);

   *loc_time = *ltm;
   return timer;
}


/*
 ***************************************************************************
 * Find number of processors used on the machine
 * (0 means one proc, 1 means two proc, etc.)
 * As far as I know, there are two possibilities for this:
 * 1) Use /proc/stat or 2) Use /proc/cpuinfo
 * (I haven't heard of a better method to guess it...)
 * See kernel sources:
 * 2.4: linux/fs/proc/proc_misc.c: kstat_read_proc()
 * 2.6: linux/fs/proc/proc_misc.c: show_stat()
 ***************************************************************************
 */
int get_cpu_nr(int *cpu_nr, unsigned int max_nr_cpus)
{
   FILE *statfp;
   char line[16];
   int proc_nb, smp_box;

   *cpu_nr = -1;

   /* Open stat file */
   if ((statfp = fopen(STAT, "r")) == NULL) {
      fprintf(stderr, _("Cannot open %s: %s\n"), STAT, strerror(errno));
      exit(1);
   }

   while (fgets(line, 16, statfp) != NULL) {

      if (strncmp(line, "cpu ", 4) && !strncmp(line, "cpu", 3)) {
	 sscanf(line + 3, "%d", &proc_nb);
	 if (proc_nb > *cpu_nr)
	   *cpu_nr = proc_nb;
      }
   }

   /*
    * cpu_nr initial value: -1
    * If cpu_nr < 0 then there is only one proc.
    * If cpu_nr > 0 then this is an SMP machine.
    * If cpu_nr = 0 then there is only one proc but this is a Linux 2.2 SMP or
    * Linux 2.4 kernel.
    */
   smp_box = (*cpu_nr > 0);
   if (*cpu_nr < 0)
      *cpu_nr = 0;

   if (*cpu_nr >= max_nr_cpus) {
      fprintf(stderr, _("Cannot handle so many processors!\n"));
      exit(1);
   }

   /* Close file */
   fclose(statfp);

   return smp_box;
}


/*
 ***************************************************************************
 * Look for partitions of a given block device in /sys filesystem
 ***************************************************************************
 */
int get_dev_part_nr(char *dev_name)
{
   DIR *dir;
   struct dirent *drd;
   char dfile[MAX_PF_NAME], line[MAX_PF_NAME];
   int part = 0;

   sprintf(dfile, "%s/%s", SYSFS_BLOCK, dev_name);

   /* Open current device directory in /sys/block */
   if ((dir = opendir(dfile)) == NULL)
      return 0;

   /* Get current file entry */
   while ((drd = readdir(dir)) != NULL) {
      if (!strcmp(drd->d_name, ".") || !strcmp(drd->d_name, ".."))
	 continue;
      sprintf(line, "%s/%s/%s", dfile, drd->d_name, S_STAT);

      /* Try to guess if current entry is a directory containing a stat file */
      if (!access(line, R_OK))
	 /* Yep... */
	 part++;
   }

   /* Close directory */
   (void) closedir(dir);

   return part;
}


/*
 ***************************************************************************
 * Look for block devices present in /sys/ filesystem:
 * Check first that sysfs is mounted (done by trying to open /sys/block
 * directory), then find number of devices registered.
 ***************************************************************************
 */
int get_sysfs_dev_nr(int flags)
{
   DIR *dir;
   struct dirent *drd;
   char line[MAX_PF_NAME];
   int dev = 0;

   /* Open /sys/block directory */
   if ((dir = opendir(SYSFS_BLOCK)) == NULL)
      /* sysfs not mounted, or perhaps this is an old kernel */
      return 0;

   /* Get current file entry in /sys/block directory */
   while ((drd = readdir(dir)) != NULL) {
      if (!strcmp(drd->d_name, ".") || !strcmp(drd->d_name, ".."))
	 continue;
      sprintf(line, "%s/%s/%s", SYSFS_BLOCK, drd->d_name, S_STAT);

      /* Try to guess if current entry is a directory containing a stat file */
      if (!access(line, R_OK)) {
	 /* Yep... */
	 dev++;
	
	 if (DISPLAY_PARTITIONS(flags))
	    /* We also want the number of partitions for this device */
	    dev += get_dev_part_nr(drd->d_name);
      }
   }

    /* Close /sys/block directory */
   (void) closedir(dir);

   return dev;
}


/*
 ***************************************************************************
 * Find number of devices and partitions available in /proc/diskstats
 * See kernel sources:
 * 2.6: linux/drivers/block/genhd.c: diskstats_show()
 ***************************************************************************
 */
int get_diskstats_dev_nr(int count_part)
{
   FILE *dstatsfp;
   char line[256];
   int dev = 0, i;
   unsigned int tmp[2];

   /* Open /proc/diskstats file */
   if ((dstatsfp = fopen(DISKSTATS, "r")) == NULL)
      /* /proc/diskstats non-existent */
      return 0;

   /*
    * Counting devices and partitions is simply a matter of counting
    * the number of lines...
    */
   while (fgets(line, 256, dstatsfp) != NULL) {
      if (!count_part) {
	 i = sscanf(line, "%*d %*d %*s %*u %*u %*u %u %u",
		    &tmp[0], &tmp[1]);
	 if (i == 1)
	    /* It was a partition and not a device */
	    continue;
      }
      dev++;
   }

   /* Close file */
   fclose(dstatsfp);

   return dev;
}


/*
 ***************************************************************************
 * Find number of devices and partitions that have statistics in
 * /proc/partitions.
 * See kernel sources:
 * 2.6: linux/drivers/block/genhd.c: show_partition() (see sysfs instead)
 ***************************************************************************
 */
int get_ppartitions_dev_nr(void)
{
   FILE *ppartfp;
   char line[256];
   int dev = 0;
   unsigned int tmp;

   /* Open /proc/partitions file */
   if ((ppartfp = fopen(PPARTITIONS, "r")) == NULL)
      /* /proc/partitions non-existent */
      return 0;

   while (fgets(line, 256, ppartfp) != NULL) {
      if (sscanf(line, "%*u %*u %*u %*s %u", &tmp) == 1)
	 /*
	  * We have just read a line from /proc/partitions containing stats
	  * for a device or a partition
	  * (i.e. this is not a fake line: title, etc.)
	  */
	 dev++;
   }

   /* Close file */
   fclose(ppartfp);

   return dev;
}


/*
 ***************************************************************************
 * Find number of disk entries that are registered on the
 * "disk_io:" line in /proc/stat.
 ***************************************************************************
 */
unsigned int get_disk_io_nr(void)
{
   FILE *statfp;
   char line[8192];
   unsigned int dsk = 0;
   int pos;

   /* Open /proc/stat file */
   if ((statfp = fopen(STAT, "r")) == NULL) {
      fprintf(stderr, _("Cannot open %s: %s\n"), STAT, strerror(errno));
      exit(2);
   }

   while (fgets(line, 8192, statfp) != NULL) {

      if (!strncmp(line, "disk_io: ", 9)) {
	 for (pos = 9; pos < strlen(line) - 1; pos +=strcspn(line + pos, " ") + 1)
	    dsk++;
      }
   }

   /* Close /proc/stat file */
   fclose(statfp);

   return dsk;
}


/*
 ***************************************************************************
 * Print banner
 ***************************************************************************
 */
inline void print_gal_header(struct tm *loc_time, char *sysname, char *release, char *nodename)
{
   char cur_date[64];
   char *e;

   if (((e = getenv(TM_FMT_VAR)) != NULL) && !strcmp(e, K_ISO))
      strftime(cur_date, sizeof(cur_date), "%Y-%m-%d", loc_time);
   else
      strftime(cur_date, sizeof(cur_date), "%x", loc_time);

   printf("%s %s (%s) \t%s\n", sysname, release, nodename, cur_date);
}


#ifdef USE_NLS
/*
 ***************************************************************************
 * Init National Language Support
 ***************************************************************************
 */
void init_nls(void)
{
   setlocale(LC_MESSAGES, "");
   setlocale(LC_CTYPE, "");
   setlocale(LC_TIME, "");
   setlocale(LC_NUMERIC, "");

   bindtextdomain(PACKAGE, LOCALEDIR);
   textdomain(PACKAGE);
}
#endif


/*
 ***************************************************************************
 * Get window height (number of lines)
 ***************************************************************************
 */
int get_win_height(void)
{
   struct winsize win;
   /*
    * This default value will be used whenever STDOUT
    * is redirected to a pipe or a file
    */
   int rows = 3600 * 24;


   if ((ioctl(STDOUT_FILENO, TIOCGWINSZ, &win) != -1) && (win.ws_row > 2))
      rows = win.ws_row - 2;

   return rows;
}


/*
 ***************************************************************************
 * Remove /dev from path name
 ***************************************************************************
 */
char *device_name(char *name)
{
   if (!strncmp(name, "/dev/", 5))
      return name + 5;

   return name;
}


/*
 ***************************************************************************
 * Get page shift in kB
 ***************************************************************************
 */
int get_kb_shift(void)
{
   int shift = 0;
   int size;

   size = PAGE_SIZE >> 10; /* Assume that a page has a minimum size of 1 kB */
   while (size > 1) {
      shift++;
      size >>= 1;
   }

   return shift;
}


/*
 ***************************************************************************
 * Handle overflow conditions properly for counters which are read as
 * unsigned long long, but which can be unsigned long long or
 * unsigned long only depending on the kernel version used.
 * @value1 and @value2 being two values successively read for this
 * counter, if @value2 < @value1 and @value1 <= 0xffffffff, then we can
 * assume that the counter's type was unsigned long and has overflown, and
 * so the difference @value2 - @value1 must be casted to this type.
 ***************************************************************************
 */
double ll_sp_value(unsigned long long value1, unsigned long long value2,
		   unsigned long long itv)
{
   if ((value2 < value1) && (value1 <= 0xffffffff))
      /* Counter's type was unsigned long and has overflown */
      return ((double) ((value2 - value1) & 0xffffffff)) / itv * 100;
   else
      return SP_VALUE(value1, value2, itv);
}

double ll_s_value(unsigned long long value1, unsigned long long value2,
		  unsigned long long itv)
{
   if ((value2 < value1) && (value1 <= 0xffffffff))
      /* Counter's type was unsigned long and has overflown */
      return ((double) ((value2 - value1) & 0xffffffff)) / itv * HZ;
   else
      return S_VALUE(value1, value2, itv);
}

/*
 *****************************************************************************
 * ioconf handling code
 *  Provides:
 *    ioc_init() - reads IOCONF file (defined below) into a data structure
 *                  that's static to this module, returning 1 on success,
 *                  0, otherwise.
 *
 *
 */

#include <linux/major.h>

#define IOC_NAMELEN 31
#define IOC_DESCLEN 63
#define IOC_DEVLEN  47
#define IOC_MAXMINOR 255
#define IOC_LINESIZ 255
#define IOC_PARTLEN 7
#define IOC_FMTLEN 15
#ifndef IOCONF
#define IOCONF "/etc/sysconfig/sysstat.ioconf"
#endif
#define IS_WHOLE(maj,min) ((min%ioconf[maj]->blkp->pcount) == 0)

static unsigned int ioc_parsed = 0;

struct blk_config {
  char name[IOC_NAMELEN+1];	/* device basename */
  char cfmt[IOC_FMTLEN+1];	/* controller format string */
  char dfmt[IOC_FMTLEN+1];	/* disk format string */
  char pfmt[IOC_FMTLEN+1];	/* partition format string */
  /* ctrlno is in the ioc_entry */
  unsigned int ctrl_explicit;	/* use "cN" in name */
  unsigned int dcount;		/* number of devices handled by this major */
  unsigned int pcount;		/* partitions per device */
  char desc[IOC_DESCLEN+1];
  /* disk info unit # conversion function */
  char *(*cconv)(unsigned int);

  /* extension properties (all this for initrd?) */
  char ext_name[IOC_NAMELEN+1];
  unsigned int ext;		/* flag - this is an extension record */
  unsigned int ext_minor;	/* which minor does this apply to */
};

struct ioc_entry {
  int live;			/* is this a Direct entry? */
  unsigned int ctrlno;		/* controller number */
  char *desc;			/* (dynamic) per-controller description */
  struct blk_config *blkp;	/* the real info, may be a shared ref */
};

static struct ioc_entry *ioconf[MAX_BLKDEV+1];

static void
ioc_free(void) {

  unsigned int i;
  struct ioc_entry **p;

  /* take out all of the references first */
  for( i = 0, p = ioconf; i < MAX_BLKDEV; ++i, ++p ) {
    if( *p == NULL || (*p)->live ) continue;
    
    if( (*p)->desc != (*p)->blkp->desc ) {
      /* not a shared description */
      free( (*p)->desc );
    }
    free( *p );
    *p = NULL;
  }

  /* now the live ones */
  for( i = 0, p = ioconf; i < MAX_BLKDEV; ++i, ++p ) {
    if( *p == NULL ) continue;
    free( (*p)->blkp );
    free( *p );
    *p = NULL;
  }
}


/*
 * ioc_conv -
 *   turn a number into a string in radix radix using symbol
 *   set (and ordering) syms.  Use nozero to generate strings
 *   in which the number system uses a single sym for the
 *   radix value (not 2, like decimal) and adds a column only
 *   at radix+1.  If decimal were like this:
 *
 *   (no zero) 1 2 3 4 5 6 7 8 9 0 11 12 13 14 15 16 17 18 19 10 ...
 */

static char *
ioc_conv( int radix, int nozero, const char *syms, unsigned int val ) {

  static char out[16];
  char *p;
  int j;

  *(p = out+16) = '\0';

  val += nozero;

  if( val == 0 ) {
    if( !nozero ) {
      *--p = '0';
    }
    return(p); 		/* empty string if nozero radix gets val == 0 */
  }

  while( val > 0 ) {
    *--p = syms[j = val%radix];
    val /= radix;
    if( nozero && (j == 0) ) {
      /* comp for 10 in nozero bases */
      --val;
    }
  }
  return(p);
}


char *
ioc_ito10( unsigned int n ) {

  return( ioc_conv( 10, 0, "0123456789", n ) );
}

char *
ioc_ito26( unsigned int n ) {
  
  return( ioc_conv( 26, 1, "zabcdefghijklmnopqrstuvwxy", n ) );
}



/*
 *  ioc_name() -
 *     generate a name from a maj,min pair
 *
 *  returns NULL if major or minor are out of range
 *  otherwise returns a pointer to a static string containing
 *   the generated name
 */

char *
ioc_name( unsigned int major, unsigned int minor ) {

  static char name[IOC_DEVLEN+1];
  struct ioc_entry *p;
  int base, offset;

  if( MAX_BLKDEV <= major || IOC_MAXMINOR <= minor ) {
    return(NULL);
  }

  if( !ioc_parsed && !ioc_init() ) return( NULL );

  p = ioconf[major];

  /* invalid major or minor numbers? */
  if( p == NULL || minor >= (p->blkp->dcount * p->blkp->pcount) ) {
    /* that minor test is only there for IDE-style devices
     *  that have no minors over 128
     */
    strcpy( name, "nodev" );
    return(name);
  }

  /* is this an extension record? */
  if( p->blkp->ext && p->blkp->ext_minor == minor ) {
    strcpy( name, p->blkp->ext_name );
    return(name);
  }
  
  /* OK.  we're doing an actual device name... */

  /* assemble base + optional controller info
   *  this is of course too clever by half
   *  the parser has already cooked cfmt, dfmt to make this easy
   *   (we parse once but may generate lots of names)
   */
  base = p->ctrlno * p->blkp->dcount;
  offset = minor / p->blkp->pcount;
  if( !p->blkp->ctrl_explicit ) offset += base;

  /* these sprintfs can't be coalesced because the first might
   *  ignore its first arg
   */
  sprintf( name, p->blkp->cfmt, p->ctrlno );
  sprintf( name + strlen(name), p->blkp->dfmt, p->blkp->cconv( offset ) );
  
  if( !IS_WHOLE( major, minor ) ) {
    /* tack on partition info, format string cooked (curried?) by
    *   the parser
    */
    sprintf( name + strlen(name), p->blkp->pfmt, minor % p->blkp->pcount );
  }
  return( name );
}



/* when is C going to get templates?
 */
#define IOC_ALLOC(P,TYPE,SIZE)                                           \
    do {                                                                 \
      if( P == NULL ) {                                                  \
        P = (TYPE *) malloc( SIZE );                                     \
        if( P == NULL ) {                                                \
	  fprintf( stderr, "%s: memory exhaustion!? at record #%u\n",    \
		 __FUNCTION__, count );                                  \
	  ioc_free();                                                    \
	  return( 0 );                                                   \
    }                                                                \
      }                                                                  \
    }                                                                    \
    while(0)

/* that dummy while allows ';' on the line that invokes the macro...
 */


/*
 *************************
 * ioc_init() - internalize the ioconf file
 *
 * given:    void
 * does:     parses IOCONF into ioconf, an array of ioc_entry *
 *           only entries having lines in IOCONF will have valid pointers
 * return:   1 on success
 *           0 on failure
 */

int
ioc_init(void) {

  FILE              *fp;
  unsigned int      i, major, indirect, count = 0;
  char              buf[IOC_LINESIZ+1];
  char              cfmt[IOC_FMTLEN+1];
  char              dfmt[IOC_FMTLEN+1];
  char              pfmt[IOC_FMTLEN+1];
  char              desc[IOC_DESCLEN+1];
  struct ioc_entry  *iocp = NULL;
  struct blk_config *blkp = NULL;

  if( (fp = fopen( IOCONF, "r" )) == NULL ) {
    fprintf( stderr, "%s: can't open %s\n", __FUNCTION__, IOCONF );
  return(0);
  }

  while( fgets( buf, IOC_LINESIZ, fp ) ) {

    if( *buf == '#' || *buf == '\n' ) continue;

    /* preallocate some (probably) needed data structures
     */
    IOC_ALLOC(blkp, struct blk_config, sizeof(struct blk_config));
    IOC_ALLOC(iocp, struct ioc_entry, sizeof(struct ioc_entry));
    memset( blkp, 0, sizeof(struct blk_config) );
    memset( iocp, 0, sizeof(struct ioc_entry) );

    i = sscanf( buf, "%u:%u:%u:%s",
		&major, &indirect, &iocp->ctrlno, desc );

    if( i != 4 ) {
      i = sscanf( buf, "%u:%u:%u",
		  &major, &indirect, &iocp->ctrlno );
    }

    if( (i == 3) || (i == 4) ) {
      /* indirect record */
      if( indirect == 0 ) {
	/* conventional usage for unsupported device */
	continue;
      }
      if( indirect >= MAX_BLKDEV ) {
	fprintf( stderr, "%s: indirect major #%u out of range\n",
		 __FUNCTION__, indirect );
	continue;
      }
      if( ioconf[indirect] == NULL ) {
	fprintf( stderr,
		 "%s: indirect record '%u:%u:%u:...'"
		 " references not yet seen major %u\n",
		 __FUNCTION__, major, indirect, iocp->ctrlno, major );
	continue;
      }
      /* cool. point this device at it's referent
       */
      /* skip last : (last field my be empty...)
       *   if it was empty and : was in the sscanf spec
       *   we'd only see 3 fields...
       */
      if( i == 3 ) {
	/* reference the mothership */
	iocp->desc = ioconf[indirect]->blkp->desc;
      }
      else {
	IOC_ALLOC( iocp->desc, char, IOC_DESCLEN+1 );
	strncpy( iocp->desc, desc, IOC_DESCLEN );
      }
      ioconf[major] = iocp;
      iocp->blkp = ioconf[indirect]->blkp;
      iocp->live = 0;
      iocp = NULL;
      continue;
      /* all done with indirect record */
    }

    /* maybe it's a full record? */
      
    i = sscanf( buf, "%u:%[^:]:%[^:]:%d:%[^:]:%u:%[^:]:%u:%s",
		&major, blkp->name,
		cfmt, &iocp->ctrlno,
		dfmt, &blkp->dcount,
		pfmt, &blkp->pcount,
		desc);
    
    if( i != 9 ) {
      fprintf( stderr, "%s: malformed %d field record: %s",
	       __FUNCTION__, i, buf );
      continue;
    }

    /* this is a full-fledged direct record */


    if( major == 0 || major >= MAX_BLKDEV ) {
      fprintf( stderr, "%s: major #%u out of range\n",
	       __FUNCTION__, major );
      continue;
    }

    /* is this an exception record? */
    if( *cfmt == 'x' ) {
      struct blk_config *xblkp;

      /* device has an aliased minor
       *  for now we only support on exception per major
       *  (catering to initrd: (1,250))
       */
      if( ioconf[major] == NULL ) {
	fprintf( stderr, "%s: type 'x' record for"
		 " major #%u must follow the base record - ignored\n",
		 __FUNCTION__, major );
	continue;
      }
      xblkp = ioconf[major]->blkp;
      if( xblkp->ext ) {
	/* enforce one minor exception per major policy
	 *  note: this applies to each major number and
	 *   all of it's indirect (short form) majors
	 */
	fprintf( stderr, "%s: duplicate 'x' record for"
		" major #%u - ignored\ninput line: %s\n",
		__FUNCTION__, major, buf );
	continue;
      }
      /* decorate the base major struct with the
       *  exception info
       */
      xblkp->ext_minor = iocp->ctrlno;
      strcpy( xblkp->ext_name, blkp->name );
      xblkp->ext = 1;
      continue;
    }

    /* preformat the sprintf format strings for generating
     *  c-d-p info in ioc_name()
     */

    /* basename of device + provided string + controller # */
    if( *cfmt == '*' ) {
      strcpy( blkp->cfmt, blkp->name );
    }
    else {
      sprintf( blkp->cfmt, "%s%s%%d", blkp->name, cfmt );
      ++(blkp->ctrl_explicit);
    }
    
    /* Disk */
    *blkp->dfmt = '\0';
    switch( *dfmt ) {
    case 'a':
      blkp->cconv = ioc_ito26;
      strcpy( blkp->dfmt, "%s" );
      break;
    case '%':
      strcpy( blkp->dfmt, dfmt + 1 );
    case 'd':
      blkp->cconv = ioc_ito10;
      strcat( blkp->dfmt, "%s" );
      break;
    }

    /* Partition */
    sprintf( blkp->pfmt, "%s%%d", (*pfmt == '*') ? "" : pfmt ); 

    /* we're good to go.
     *  stuff the ioc_entry and ref it
     */
    iocp->live = 1;
    iocp->blkp = blkp;
    iocp->desc = NULL;
    ioconf[major] = iocp;
    strncpy( blkp->desc, desc, IOC_DESCLEN );
    blkp = NULL; iocp = NULL;
    ++count;
  }
  fclose(fp);

  /* these will become leaks if we ever 'continue'
   *  after IOC_ALLOC( blkp->desc ... ).
   * Right Now, we don't
   */
  if( blkp != NULL ) free(blkp);
  if( iocp != NULL ) free(iocp);

  /* Indicate that ioconf file has been parsed */ 
  ioc_parsed = 1; 

  return( count );
}   


/**********************************************************************
 * int ioc_iswhole(major,minor)
 *
 *  predicate.  returns 1 if dev (major,minor) is a whole disk device
 *              returns 0 otherwise
 */

int
ioc_iswhole( unsigned int major, unsigned int minor )
{
  if( !ioc_parsed && !ioc_init() ) {
    return(0);
  }
  if( major >= MAX_BLKDEV ) {
    /* later: handle Linux long major numbers here
     * now: this is an error
     */
    return(0);
  }
  if( ioconf[major] == NULL ) {
    /* device not registered */
    return(0);
  }

  return( IS_WHOLE( major, minor ) );
}

/*  
 *************************************************************************** 
 * Find number of NFS-mounted points that are registered in 
 * in /proc/self/mount. 
 *************************************************************************** 
 */ 
int get_nfs_mount_nr(void) 
{ 
   FILE *fp; 
   char line[8192],pos[8192]; 
   char type_name[10]; 
   unsigned int nfs = 0; 
    
   if ((fp = fopen(NFSMOUNTSTATS, "r")) == NULL) { 
      fprintf(stderr, _("Cannot open %s: %s\n"), NFSMOUNTSTATS, strerror(errno)); 
      exit(2); 
   } 
   while (fgets(line, 8192, fp) != NULL) { 
      if ((strstr(line,"mounted")) && (strstr(line,"on")) && (strstr(line,"with")) && (strstr(line,"fstype"))) { 
         sscanf(strstr(line,"fstype")+6,"%10s", type_name);  
         if ((!strncmp(type_name,"nfs",3)) && (strncmp(type_name,"nfsd",4))) { 
            nfs ++;      
        }  
      } 
   } 
   fclose(fp); 
   return nfs; 
} 
 
 
